#include "RdbJoin.h"


int RdbJoin::Arity() const
{
    int rels = relList.Size();
    int arity = 0;
    for( int i = 0 ; i < rels ; i++ )
      arity += relList[i]->Arity();
    return arity;
}


String RdbJoin::Schema() const
{
    String s(relList[0]->Schema());
    int rels = RelCount();
    for( int i = 1 ; i < rels ; i++ )
      s += String(",") + relList[i]->Schema();
    // should include join info...
    return s;
}


void RdbJoin::ColNumberToRelCol(int colNum, int& rel, int& col) const
{
    rel = 0;
    col = 0;
    int relArity = relList[0]->Arity();
    while( colNum >= relArity ) {
	colNum -= relArity;
	rel++;
	assert( rel < relList.Size() );
	relArity = relList[rel]->Arity();
    }
    col = colNum;
}


String RdbJoin::QualifiedColName(int rel,int col) const
{
    return relList[rel]->RelName() + "." + relList[rel]->ColName(col);
}


String RdbJoin::QualifiedColName(int colNum) const
{
    int rel, col;
    ColNumberToRelCol(colNum,rel,col);
    return QualifiedColName(rel,col);
}


String RdbJoin::VarQualifiedColName(int rel,int col) const
{
    return String("R") + long(rel) + "." + relList[rel]->ColName(col);
}


String RdbJoin::VarQualifiedColName(int colNum) const
{
    int rel, col;
    ColNumberToRelCol(colNum,rel,col);
    return VarQualifiedColName(rel,col);
}


String RdbJoin::VarJoinString(int j) const
{
    RdbJoinCondition& joinInfo = joinCond[j];
    int rel1 = joinInfo.Rel1();
    int rel2 = joinInfo.Rel2();
    int col1 = joinInfo.Col1();
    int col2 = joinInfo.Col2();
    return VarQualifiedColName(rel1,col1) + " = " 
      + VarQualifiedColName(rel2,col2);
}
