#ifndef _RDBRELATION_H_
#define _RDBRELATION_H_

// A relation is a backend relation name, collection of attribute 
// names (columns), and database in which it resides.

#include "RdbDatabase.h"
#include "RdbError.h"
#include "List.h"
#include "String.h"


typedef List<String> RdbColNameList;

class RdbRelation
: public RdbError
{
  public:

    RdbRelation(RdbDatabase& newDb, String& newRelName, 
		RdbColNameList& newColNames );
				// map to relation newRelName 
				// (attrs newColNames) in database newDb

    ~RdbRelation();

    RdbSystem& System() const;	// System associated with this relation

    RdbDatabase& Database() const; // db associated with this relation

    int Arity() const;		// how many columns?

    const String& RelName() const; // name of the relation in backend

    const String& ColName(int i) const; // column name in backend

    String Schema() const;	// relName(col1,,coln)

    String SystemName() const;	// system that this db is in

    String DatabaseName() const; // database that this is in

    RdbInsert* NewInsert();	// create an insert stmt on this relation

    RdbDelete* NewDelete(const BitVector& adorn); 
				// create a delete stmt on this relation

  private:

    RdbDatabase& database;	// the database this is in

    String relName;		// name of the rel in the backend

    RdbColNameList colNames;	// names of the attrs in the backend

    RdbRelation(const RdbRelation&); // not allowed

    RdbRelation& operator =(const RdbRelation&); // not allowed
};

#include "RdbRelation.i"

#endif // _RDBRELATION_H_
