# line 1 "RdbIngresQuery.scc"	

#include "RdbIngresQuery.h"
#include "Log.h"

#ifndef _SQLCA_H_
#define _SQLCA_H_
# include "/ingres/r6/ingres/files/eqsqlca.h"
    extern IISQLCA sqlca;   /* SQL Communications Area */
#endif
#ifndef _SQLDA_H_
#define _SQLDA_H_
# include "/ingres/r6/ingres/files/eqsqlda.h"
#endif
#include "IngresProtos.h"


// shared variables
unsigned RdbIngresQuery::queryNum = 0;
unsigned RdbIngresQuery::cursorNum = 0;


RdbIngresQuery::RdbIngresQuery(const RdbJoin& newJoin, 
			       const BitVector& newAdorn)
: RdbQuery(newJoin,newAdorn)
{ 
    id = String("qry") + queryNum++; 
    CreateStmt();
    Describe();
}


RdbError::ErrorCode RdbIngresQuery::VOpen(RdbCursor& cursor)
{
  char *cname;
  char *qname = id;

    SetCursorId(cursor, new String( String("csr") + cursorNum++ ));
    cname = GetCursorId(cursor);
    if( !IsReady() )
      Prepare();
    Bind( cursor.Bindings() );
    IISQLDA* whereSqlda = whereDA;
#   ifdef DO_LOGGING
    {
	LOG("Opening cursor "<<cname<<" for query "<<qname<<'\n');
	for( int i = 0 ; i < whereDA.Arity() ; i++ )
	  LOG("    col #"<<i<<" bound to "<<whereDA[i]<<'\n');
    }
#   endif
# line 52 "RdbIngresQuery.scc"	/* open */
  {
    IIsqInit(&sqlca);
    IIcsOpen(cname,(int)0,(int)0);
    IIwritio((int)1,(short *)0,(int)1,(int)32,(int)0,qname);
    IIsqDaIn((int)0,whereSqlda);
    IIcsQuery(cname,(int)0,(int)0);
  }
# line 53 "RdbIngresQuery.scc"	/* host code */
    if( sqlca.sqlcode )
      return Error(SQL_ERROR);
    return NO_ERROR;
}


RdbError::ErrorCode RdbIngresQuery::VClose(RdbCursor& cursor)
{
    String& cid = GetCursorId(cursor);
  char *cname = cid;

    LOG("Closing cursor "<<cname<<endl);
# line 67 "RdbIngresQuery.scc"	/* close */
  {
    IIsqInit(&sqlca);
    IIcsClose(cname,(int)0,(int)0);
  }
# line 68 "RdbIngresQuery.scc"	/* host code */
    delete &cid;
    if( sqlca.sqlcode )
      return Error(SQL_ERROR);
    return NO_ERROR;
}


RdbError::ErrorCode RdbIngresQuery::Describe()
{
  char *stmt = statement;
  char *qname = id;
    //resultDA.Unprepare();
    //whereDA.Unprepare();
    resultDA.Resize(Arity());
    IISQLDA* sqlda = resultDA;
    LOG("Describing "<<stmt<<'\n');
    //xec sql describe :stmt into sqlda;
# line 87 "RdbIngresQuery.scc"	/* prepare */
  {
    IIsqInit(&sqlca);
    IIsqPrepare((int)0,qname,sqlda,(int)0,stmt);
  }
# line 88 "RdbIngresQuery.scc"	/* host code */
    if( sqlca.sqlcode )
      return Error(SQL_ERROR);
    resultDA.Prepare();
    whereDA.Prepare( resultDA, Adornment() );
    assert(Arity() == resultDA.Arity());
    ready = TRUE;
    return NO_ERROR;
}


RdbError::ErrorCode RdbIngresQuery::Bind(const RdbItemList& bindings)
{
    int n = bindings.Size();
    if( n != whereDA.Arity() )
      return Error(BAD_ARITY);
    for( int i = 0; i < n ; i++ ) {
	// should probably check for compatiblilty
	whereDA.SetItem( i, bindings[i] );
	LOG("binding it to: "<<bindings[i]<<'\n');
    }
    return NO_ERROR;
}


RdbError::ErrorCode RdbIngresQuery::VGetNext(RdbCursor& cursor,
					     RdbItemList& result)
{
  char *cname = GetCursorId(cursor);
    IISQLDA* sqlda = resultDA;

    LOG("fetching on cursor "<<cname<<'\n');
# line 121 "RdbIngresQuery.scc"	/* fetch */
  {
    IIsqInit(&sqlca);
    if (IIcsRetrieve(cname,(int)0,(int)0) != (int)0) {
      IIcsDaGet((int)0,sqlda);
      IIcsERetrieve();
    } /* IIcsRetrieve */
  }
# line 122 "RdbIngresQuery.scc"	/* host code */
    if( sqlca.sqlcode == 100 )
      return SQL_EOF;
    if( sqlca.sqlcode )
      return Error(SQL_ERROR);
    result.DeleteAll();
    for( int i = 0 ; i < resultDA.Arity() ; i++ )
      result.Append( resultDA[i] );
    return NO_ERROR;
}


