// *************************************************************************
//    CompileRule.C
// *************************************************************************

#include <stdio.h>
#include "CompileRule.h"

#include "Literal.h"
#include "Unparse.h"
#include "SymTable.h"
#include "Database.h"
#include "BuiltinRelation.h"
#include "RelationMisc.h"
#include "Annotations.h"
#include "Predicate.h"
#include "BindTime.h"
#include "Debug.h"

// *************************************************************************
// WARNING: The following routine empties the varlist that is passed to it.
void Cor_create_an_annotation(Cor_PredAnnotations *&annos, Cor_Literal *lit,
    Cor_IntStack *var_list)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

        // count of number of variables in var_list
    int var_count = var_list->count();
        // count of number of constants in lit->args
    int const_count = 0;

    FOR_EACH_ARG(a, lit->args) {
        if (a->hash(NULL) != VarHashValue)
            const_count++;
    } END_EACH_ARG;

    // no point generating an index with all arguments free !!
    if (!var_count && !const_count)
        return;

    Cor_ArgList *arglist = new Cor_ArgList(var_count);

      // The order is important here.
    Cor_CArg varg;
    while (! var_list->is_empty()) {
        varg.set_var(var_list->pop());
        (*arglist)[--var_count]= varg;
    }

    Cor_CArg name;

    if (lit->kind == COR_MAGIC) {

        char *buf = new char [lit->pred.symlen()+10];
        Cor_sprint_magic_name(buf,lit->pred.symbol(),"");
        name = Cor_CArg(buf);
        delete [] buf;

    } else if (lit->kind == COR_DONE_MAGIC) {

        char *buf = new char [lit->pred.symlen()+15];
        strcpy(buf, "done_");
        char *buf1 = buf + strlen(buf);
        Cor_sprint_magic_name(buf1,lit->pred.symbol(),"");
        name = Cor_CArg(buf);
        delete [] buf;

    } else {
        name = lit->pred;
    }

    Cor_add_make_index_annotation(annos,name, &(lit->adorn),
                        &(lit->args), arglist);
    delete arglist;
}

// *************************************************************************
Cor_IntStack *Cor_vars_between(Cor_Clause *adrule, int start, int end,
        Cor_IntStack *var_stk, int *predOrders )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    ASSERT(var_stk);

    for (i=start; i <= end; i++) {
        if (predOrders)
            Cor_build_var_list(adrule->preds[predOrders[i]], var_stk);
        else
            Cor_build_var_list(adrule->preds[i], var_stk);
    }

    return(var_stk);
}

// *************************************************************************
int Cor_is_a_builtin(Cor_CArg& name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_BuiltinRelation *bi_rel = (Cor_BuiltinRelation *)
            Cor_BuiltinDB->RelationTable->lookup(name);

        // Used to check for is_printable_builtin(), but not anymore.
    if (bi_rel)
        return(1);

    return(0);
}

// *************************************************************************
//  This funtion tries to find the literal in the rule that is best suited to
//  be placed immediately after 'start' in the join order.
// *************************************************************************
static int find_best_next_literal(Cor_Clause *rulePtr, int *predOrders,
    Cor_IntStack *leading_vars, int start, int end)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    int i;
    int best = start;
    int bestval = -1;

    if (Cor_is_a_builtin(rulePtr->preds[predOrders[start]]->pred))
        return(start);

    Cor_IntStack *overlap_vars = new Cor_IntStack();

    for (i = start; i <= end; i++) {
        if (Cor_is_a_builtin(rulePtr->preds[predOrders[i]]->pred))
            continue;
        Cor_intersect_var_lists(
            Cor_build_var_list(rulePtr->preds[predOrders[i]],overlap_vars),
            leading_vars);
        if (overlap_vars->count() > bestval) {
            bestval = overlap_vars->count();
            best = i;
        }

        overlap_vars->empty();
    }

    delete overlap_vars;

    return(best);
}

// *************************************************************************
void Cor_generateIndexes(Cor_SemiNaive *snrule, Cor_PredAnnotations *&annos)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i, j, k;
    Cor_IntStack *leading_vars = new Cor_IntStack();
    Cor_IntStack *indexing_vars = new Cor_IntStack();

#    define RULEPTR snrule->rInfo->rule
#    define PREDINDEX(i) snrule->predOrders[i]

    // The following code indexes predicates in the rule
    for (i=0; i < RULEPTR->num_preds; i++) {

        Cor_vars_between(RULEPTR,0,i-1,leading_vars,snrule->predOrders);
        if (i < snrule->first) {
            j = find_best_next_literal(RULEPTR, snrule->predOrders,
                leading_vars, i, snrule->first);
            k = snrule->predOrders[j];
            while (j > i) {
                snrule->predOrders[j] = snrule->predOrders[j-1];
                j--;
            }
            snrule->predOrders[j] = k;
        }

        // no point building an index on an arithmetic predicate
        if (!Cor_is_a_builtin(RULEPTR->preds[PREDINDEX(i)]->pred)) {
            Cor_intersect_var_lists( Cor_build_var_list(
                   RULEPTR->preds[PREDINDEX(i)],indexing_vars), leading_vars);
            Cor_create_an_annotation( annos,
                RULEPTR->preds[PREDINDEX(i)], indexing_vars);
        }

        leading_vars->empty();
        indexing_vars->empty();
    }

    delete leading_vars;
    delete indexing_vars;

    // if there is a grouping argument in the head, create an index on the
    // the head predicate that is bound on all arguments except the grouping
    // argument

    if (RULEPTR->head->args.has_grouping_args()) {
        char *buf = new char[RULEPTR->head->args.count() + 1];
        for (i=0; i< RULEPTR->head->args.count(); i++) {
            if (RULEPTR->head->args[i].is_grouping())
                buf[i] = 'f';
            else
                buf[i] = 'b';
        }
        buf[i] = '\0';
        Cor_ArgList *new_arglist = new Cor_ArgList(1);
        (*new_arglist)[0] = Cor_CArg(buf);
        delete [] buf;

        Cor_add_make_index_annotation(annos, RULEPTR->head->pred,
            &(RULEPTR->head->adorn), new_arglist, NULL);
        delete new_arglist;
    }

#    undef RULEPTR
#    undef PREDINDEX
}

// *************************************************************************
void Cor_initPredOrders(Cor_SemiNaive *sn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;

    if (sn->rInfo->num_literals) {
        sn->predOrders[0] = sn->first;
        for (j = 1; j < sn->rInfo->num_literals; j++) {
            if (j <= sn->first)
                sn->predOrders[j] = j-1;
            else
                sn->predOrders[j] = j;
        }
     }
}

// *************************************************************************
int initFailureBacktrack(Cor_SemiNaive *sn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_BindTime bta;

    if (!bta.Analyze(sn->rInfo, sn->predOrders))
        bta.FixOrder(sn->predOrders);

    //cout << "Setting Unify Functions (for RuleInfo: "<<sn->rInfo<<").\n";
    bta.set_UnifyFunctions(sn->rInfo, sn->unify_funct, sn->predOrders);

    /*
    if (sn->predOrders) {
        cout << "predOrders exists:\n";
        for (int j=0; j < sn->rInfo->num_literals; j++)
            cout << sn->predOrders[j] << " ";
        cout << endl;
    }

    bta.show_raw_info();
    */

    bta.get_firstFailure_BT_Pt(sn->firstFailureBacktrack);
    bta.get_nextFailure_BT_Pt(sn->nextFailureBacktrack);

    return(bta.get_Success_BT_Pt());
}

// *************************************************************************
Cor_RuleInfo *Cor_create_rule_info(Cor_Table *F, Cor_Clause *rule)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int num_preds = rule->num_preds;
    int num_deletes = rule->head_deletes_count;
    int ipred;

    // Variable declarations

    Cor_RuleInfo *rInfo = new Cor_RuleInfo(rule);

    //  Adding info for rInfo->arg_list, and rInfo->relations.
    //  and space to select which Unify function to use.

    for (ipred = 0; ipred < num_preds; ipred++) {
        rInfo->arg_list[ipred] = rule->preds[ipred]->args;
    }

    rInfo->arg_list[num_preds] = rule->head->args;
    for (ipred=0; ipred < num_deletes; ipred++) {
        rInfo->arg_list[num_preds+1+ipred] =
            rule->head_deletes[ipred]->args;
    }

    // If the predicate is negated, a negative value is placed in the
    // rInfo->relations[] field. This info is used by SemiNaive::evaluate()
    for (ipred = 0; ipred < num_preds; ipred++) {
        if (rule->preds[ipred]->negated) {
            rInfo->relations[ipred]= -2 - (rInfo->offsets[ipred] =
                Cor_get_offset(F, rule->preds[ipred]));
        } else {
            rInfo->relations[ipred]=
                (rInfo->offsets[ipred] = Cor_get_offset(F, rule->preds[ipred]));
        }
    }

    if (rule->head->negated) {
        rInfo->relations[num_preds] = -2 - (rInfo->offsets[num_preds] =
            Cor_get_offset(F, rule->head));
    } else {
        rInfo->relations[num_preds] =
            (rInfo->offsets[num_preds] = Cor_get_offset(F, rule->head));
    }

    for (ipred = 0; ipred < num_deletes; ipred++) {
        rInfo->relations[num_preds+1+ipred] =
            -2 - (rInfo->offsets[num_preds+1+ipred] =
            Cor_get_offset(F, rule->head_deletes[ipred]));
    }

    rInfo->relations[num_preds+num_deletes+1] = -1;
    rInfo->offsets[num_preds+num_deletes+1] = -1;

    return(rInfo);
}

// *************************************************************************
void Cor_compile_nonrec_rule(Cor_SemiNaive &sn, Cor_Clause *clause,
        Cor_RuleInfo *rInfo)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    sn.first = 0;
    sn.recursive = 0;
    sn.Add_Rule_Info(rInfo, clause);

    for (int jpred = 0; jpred < clause->num_preds; jpred++)
        sn.markOffsets[2*jpred] = sn.markOffsets[2*jpred+1] = -1;
}

// *************************************************************************
