// *************************************************************************
//    Execute.C
// *************************************************************************

#include "Execute.h"

#include "ArgList.h"
#include "StorageRelation.h"
#include "BindEnv.h"
#include "ModuleInfo.h"
#include "SccInfo.h"
#include "Globals.h"
#include "HashSimpleRelation.h"
#include "Compile.h"
#include "Suspend.h"
#include "Suspension.h"
#include "Debug.h"

// *************************************************************************

#ifdef WITH_PERSISTENCE
extern Cor_StorageRelation *make_pers_relation(Cor_CArg&, int arity,
    int delta_index, int set_subsum, Cor_ArgList *a_list);
#define MAKE_PERS_REL_MACRO(a,b,c,d,e) make_pers_relation(a,b,c,d,e)
#else
#define MAKE_PERS_REL_MACRO(a,b,c,d,e) Cor_make_local_relation(a,b,c,d)
#endif

// *************************************************************************

// points to current moduleInfo defaults
Cor_ModuleInfo *Cor_CurrModuleInfo = NULL;

// *************************************************************************
void Cor_PrintSccRecursiveRules(Cor_SccInfo *scc, int start, ostream *outf)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;

    *outf << "\n[Iterating over recursive rules ...]";
    *outf << "\n---------------------------------------\n";

    for (int i = start; i < scc->numSNRules ; i++) {
#ifdef    DO_PROFILE
        *outf << scc->sn_rules[i].rInfo->pInfo.rule << endl;
#endif
        *outf << "Predord :: ";
        for (j = 0; j < scc->sn_rules[i].rInfo->num_literals; j++)
            *outf << scc->sn_rules[i].predOrders[j] << " ";
        *outf << "\n1st bt :: ";
        for (j = 0; j < scc->sn_rules[i].rInfo->num_literals; j++)
            *outf << scc->sn_rules[i].firstFailureBacktrack[j] << " ";
        *outf << "\nnext bt :: ";
        for (j = 0; j < scc->sn_rules[i].rInfo->num_literals; j++)
            *outf << scc->sn_rules[i].nextFailureBacktrack[j] << " ";
        *outf << "\n\n";
    }

    *outf << "---------------------------------------\n\n";
    *outf << flush;
}

// *************************************************************************
void Cor_PrintSccNonRecursiveRules(Cor_SccInfo *scc, ostream *outf)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *outf << "\n[Applying non-recursive rules ...]";
    *outf << "\n---------------------------------------\n";

    for (int i = 0; i < scc->numSNRules ; i++) {
        if (scc->sn_rules[i].recursive) break;
//        if (!scc->sn_rules[i].ignore_this_rule) {
#ifdef    DO_PROFILE
            *outf << scc->sn_rules[i].rInfo->pInfo.rule << endl;
#endif
            *outf << endl;
//        }
    }

    *outf << "---------------------------------------\n\n";
}

// *************************************************************************
void Cor_EndOfScc(Cor_SccInfo *scc, int iters, ostream *outf)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;

    if (Cor_exEnv->profile_scc) {
        for (i=0; i< scc->numSNRules; i++) {
            //if (!scc->sn_rules[i].ignore_this_rule) {
#ifdef    DO_PROFILE
                scc->sn_rules[i].rInfo->pInfo.print(outf, "SCC",
                    scc->sn_rules[i].predOrders,
                    scc->sn_rules[i].rInfo->num_literals);
#endif
            //}
        }
        *outf << "\nNumber of SCC iterations = " << iters << endl;
    }
}

// *************************************************************************
static void clearStatistics(Cor_QFMaterializedModuleData *md, ostream *outf)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    DO_PROFILE
    int i, j;
#endif
    long profile_rules = ((Cor_exEnv->GlobalRelationOptions | md->rel_options) &
        REL_PROFILE_STATS);

    if (profile_rules && (Cor_exEnv->profile_module)) {
#ifdef    DO_PROFILE
        md->module_info->Cor_GlobalProfileInfo.clear();
        md->module_info->Cor_GlobalProfileInfo.set_rule_name("");
        for (i = 0; i < md->numSccs; i++) {
            for (j=0; j < md->ruleArray->numRules; j++) {
                Cor_AddProfileInfo(&(md->ruleArray->rInfoPtrs[j]->pInfo),
                    &(md->module_info->Cor_GlobalProfileInfo));
                md->ruleArray->rInfoPtrs[j]->pInfo.clear();
            }
        }
        md->module_info->Cor_GlobalProfileInfo.print(outf, "MODULE SUMMARY");
        md->module_info->Cor_GlobalProfileInfo.clear();
#endif
    } else {
#ifdef    DO_PROFILE
        for (i = 0; i < md->numSccs; i++) {
            for (j = 0; j < md->sccArray[i]->numSNRules; j++) {
                md->sccArray[i]->sn_rules[j].rInfo->pInfo.clear();
            }
        }
#endif
    }
}

// *************************************************************************
//  WARNING  - with recursion deeper than exEnv->C_max_recursion_depth_default
//  this code will complain and abort.
// *************************************************************************
Cor_StorageRelation *Cor_find_local_relation(Cor_CArg& name, int arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if ((Cor_T_Stack->count <= 0) || (Cor_T_Stack->top() == NULL))
        return NULL;

    int offset = Cor_get_offset(Cor_T_Stack->top(), name, arity, 0);
    if (offset < 0)
        return NULL;

    return (Cor_StorageRelation*)(Cor_T_Stack->top()->rels[offset].relPtr);
}

// *************************************************************************
static int init_rule_pointers(Cor_Table *F, Cor_QFModuleData *md,
    Cor_StorageRelation *magic, int reused_table, int old_magic_relation)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_StorageRelation *temp, *temprel;
    Cor_ModulePred *predicates = md->predicates->next;

    F->moduledata = md;

    for (; predicates; predicates = predicates->next) {
        Cor_CArg name = predicates->pred->name;
        int arity = predicates->pred->arity;
        int offset = predicates->offset;
        switch (predicates->kind) {
        case ExternalPredKind:
            // If reused_table is set, we should be able to avoid the
            // overhead of finding external relations all over again.
            // However if anything has changed externally (say, because
            // of a rel_close), we need to do this.
            // Perhaps we should maintain a flag and find relations again
            // only if required
            F->rels[offset].AssignRel(
                    Cor_Find_External_Relation(name, arity));
            F->rels[offset].relPtr->local_options =
                md->rel_options | Cor_exEnv->GlobalRelationOptions;
            break;

        case QueryPredKind:
            // the condition tested is not (! reused_table) as for
            // LocalPredKind below since answer relations are used
            // at the calling point even after the table is returned
            // to the free table list.
            if (predicates->disk_rel) {
                 temp = MAKE_PERS_REL_MACRO(name, arity,
                    md->module_info->IndexDeltas, predicates->check_subsum,
                    predicates->schema_def);
            } else {
                 temp = Cor_make_local_relation(name, arity,
                    md->module_info->IndexDeltas, predicates->check_subsum);
            }

            F->rels[offset].AssignRel(temp);
            temp->ordsearch = md->module_info->UseOrdSearch;
            F->rels[offset].relPtr->local_options = md->rel_options |
                Cor_exEnv->GlobalRelationOptions;
            break;

        case LocalPredKind:
            if (! reused_table) {
                // Relation is not already set up.

                if (predicates->disk_rel) {
                     temp = MAKE_PERS_REL_MACRO(name, arity,
                        md->module_info->IndexDeltas,
                        predicates->check_subsum, predicates->schema_def);
                } else {
                     temp = Cor_make_local_relation(name, arity,
                        md->module_info->IndexDeltas, predicates->check_subsum);
                }

                F->rels[offset].AssignRel(temp);
                temp->ordsearch = md->module_info->UseOrdSearch;

            } else {
                F->rels[offset].relPtr->local_options =
                    md->rel_options | Cor_exEnv->GlobalRelationOptions;
                continue;
            }

            F->rels[offset].relPtr->local_options =
                md->rel_options | Cor_exEnv->GlobalRelationOptions;
            break;

        case MagicQueryPredKind:
            if (md->module_info->UseOrdSearch && (! F->context)) {
                F->context = new Cor_Context();
            }

            if (predicates->disk_rel) {
                 F->rels[offset].AssignRel(MAKE_PERS_REL_MACRO(name, arity,
                    md->module_info->IndexDeltas, predicates->check_subsum,
                    predicates->schema_def));
            } else {
                 F->rels[offset].AssignRel(Cor_make_local_relation(name,
                        arity, md->module_info->IndexDeltas,
                        predicates->check_subsum));
            }
            old_magic_relation = 0;

            // else retain the old relation

            // NOTE: the tuples in "magic" have not been added as yet
            temprel = (Cor_StorageRelation *)(F->rels[offset].relPtr);
            temprel->ordsearch = md->module_info->UseOrdSearch;
            temprel->r_kind = COR_R_MAGIC;

            if (md->module_info->UseOrdSearch) {
                if (! temprel->for_ordsearch) {
                     temprel->for_ordsearch = new Cor_HashSimpleRelation(arity);
                }
                magic->add_all_goals(temprel->for_ordsearch, F->context,
                    temprel);
            }

            temprel->local_options =
                md->rel_options | Cor_exEnv->GlobalRelationOptions;

            break;
            // From this point on, the passed in relation "magic"
            // is useless. However, it gets reclaimed only in
            // clear_rule_pointers.

        }

        Cor_AnnoStruct *cur;

        for (cur= predicates->index_annos; cur; cur=cur->next) {
            if (cur->anno->arglist2) {
                F->rels[offset].relPtr->add_index(cur->anno->arglist1,
                    cur->num_vars, cur->anno->arglist2);
            } else {
                F->rels[offset].relPtr->add_index(cur->anno->arglist1);
                // Damn, this happens for duplicates too
                //if (F->rels[offset].relPtr->add_index(cur->anno->arglist1)
                //      == COR_U_FAIL)
                //  return(0);
            }
        }

        for (cur= predicates->aggsel_annos; cur; cur=cur->next)
            F->rels[offset].relPtr->add_agg_sel_anno(cur->anno);


        if (predicates->kind == MagicQueryPredKind) {
            if (!md->module_info->UseOrdSearch) {
                // first add the tuples in "magic" to the relation
                // there could be marks because of save modules
                magic->add_all_tuples((Cor_StorageRelation *)(F->
                    rels[offset].relPtr), 1, 1);

            } else {
                // get the fact from the Context
                int any_new_fact_in_done = 0;
                Cor_GoalNode *g_node =
                    F->context->get_unmarked_goal(any_new_fact_in_done);

                ASSERT((g_node != NULL));
                ASSERT((g_node->owner_rel == F->rels[offset].relPtr));
                g_node->owner_rel->insert_tuple(g_node);
            }
        }
    }

    return(1);
}

// *************************************************************************
static int occurs_in_module(Cor_ModulePred *predicates, Cor_CArg& name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for (predicates = predicates->next; predicates;
        predicates = predicates->next) {
        if (name.equals(predicates->pred->name))
            return 1;
    }
    return 0;
}

// *************************************************************************
static void show_rule_pointers(Cor_Table *F, Cor_QFModuleData *md)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

#ifdef    DEBUG
    Cor_ModulePred *predicates = md->predicates->next;

    cout << "-------------------------\n";
    cout << "Showing rule pointers:\n";
    for (; predicates; predicates = predicates->next) {
        Cor_CArg name = predicates->pred->name;
        int offset = predicates->offset;

        cout << "Name is: " << name << " offset is: " << offset << endl;
        Cor_StorageRelation *temprel = (Cor_StorageRelation *)
                    (F->rels[offset].relPtr);

        if (!temprel)
            continue;

        switch (predicates->kind) {

        case ExternalPredKind:
            cout << "ExternalPredKind\n";
            break;

        case QueryPredKind:
            cout << "QueryPredKind\n";
            break;

        case LocalPredKind:
            cout << "LocalPredKind\n";
            cout << "temprel: " << temprel << endl;
            if (temprel) {
            cout << "temprel->for_ordsearch: " << temprel->for_ordsearch
                 << endl;
            cout << "temprel->mp_done_reln: " << temprel->mp_done_reln
                 << endl;
            }
            break;

        case MagicQueryPredKind:
            cout << "MagicQueryPredKind\n";
            cout << "temprel: " << temprel << endl;
            if (temprel) {
                cout << "temprel->for_ordsearch: " << temprel->for_ordsearch
                     << endl;
                cout << "temprel->mp_done_reln: " << temprel->mp_done_reln
                     << endl;
            }
            break;
        }
    }

    cout << "-------------------------\n";
#endif
}

// *************************************************************************
static void remove_from_predlist(Cor_Table *F, Cor_ModulePred *preds,
                                 Cor_StorageRelation *reln)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ModulePred *curr;

    if (!reln)
        return;

    for (curr = preds->next; curr; curr = curr->next) {
        int offset = curr->offset;
        Cor_StorageRelation *temprel = (Cor_StorageRelation *)
                    (F->rels[offset].relPtr);
        if (temprel == reln)
            F->rels[offset].ClearRel();
    }
}

// *************************************************************************
static void clear_rule_pointers(Cor_Table *F, Cor_QFModuleData *md,
    Cor_StorageRelation *magic)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    // Why are we starting at the second predicate?
    // There is something special about md->predicates,
    // it holds the answer relations?
    Cor_ModulePred *predicates = md->predicates->next;
    F->moduledata = md;

    // reclaim "magic"
    if (md->module_info->UseOrdSearch) {
        magic->empty_relation(1);
    } else {
        magic->empty_relation(0);
    }
    magic->ordsearch=0;

    for (; predicates; predicates = predicates->next) {
        Cor_CArg name = predicates->pred->name;
        //int arity = predicates->pred->arity;
        int offset = predicates->offset;
        Cor_StorageRelation *temprel = (Cor_StorageRelation *)
                    (F->rels[offset].relPtr);

        if (!temprel)
            continue;

        switch (predicates->kind) {

        case ExternalPredKind:
            break;

        case QueryPredKind:
            break;

        case LocalPredKind:
        {
            if (temprel->for_ordsearch) {
                temprel->for_ordsearch->empty_relation(0);
            }

            if (temprel->mp_done_reln && !occurs_in_module(md->predicates,
                temprel->mp_done_reln->name))
            {
                // for OrdSearch
                temprel->mp_done_reln->empty_relation(1);
            }

            if (Cor_is_mp_done(temprel->name)) {
                temprel->empty_relation(1);
            } else {
                remove_from_predlist(F, predicates, temprel->mp_done_reln);
                temprel->empty_relation(1);
                Cor_reclaim_relation(temprel);
                F->rels[offset].ClearRel();
            }
            break;
        }

        case MagicQueryPredKind:
        {
            temprel->ordsearch = 0;
            // else Ordered Search will crash the next time around
            if (temprel->mp_done_reln && !occurs_in_module(md->predicates,
                            temprel->mp_done_reln->name)) {
                temprel->mp_done_reln->empty_relation(1);
            }

            if (temprel->for_ordsearch) {
                temprel->for_ordsearch->empty_relation(0);
            }

            // Reclaim some of the memory that is being wasted
            Cor_reclaim_relation(temprel);
            F->rels[offset].ClearRel();

            break;
        }
        }
    }
}

// *************************************************************************
static Cor_Table *initialize_do_query(Cor_QFModuleData *md,
    Cor_StorageRelation *magic, int old_magic_relation)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Make a copy of the table. This is to handle recursive calls
    // to do_query.  Each call should see an unmodified Table.

    Cor_Table *table_copy = new Cor_Table(md);

    if (init_rule_pointers(table_copy, md, magic, 0, old_magic_relation))
        return(table_copy);
    else
        return(NULL);
}

// *************************************************************************
void Cor_cleanup_do_query(Cor_QFModuleData *md, Cor_StorageRelation *magic,
    Cor_Table *table_copy)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

      // need to delete the tuple iterators before the relations
    table_copy->module_env->clear_tupits();

    if (md->module_info->UseOrdSearch) {
        if (table_copy->context) {
            table_copy->context->free_all();
        }
    }

    clear_rule_pointers(table_copy, md, magic);

    clearStatistics((Cor_QFMaterializedModuleData *)md, Cor_exEnv->trace_strm);
}

// *************************************************************************
//  This function is called at run time to execute a query on the module.
//  The magic fact for the query is in the magic relation magic.
//  old_magic_relation specifies if the magic relation is new or old.
//  The iterator is passed in to allow for a staggered bottom-up evaluation
//  mode, where answers at the top level of the module are returned at the
//  end of each iteration of the seminaive loop. This requires that the state
//  of the seminaive computation has to be stored across calls to do_query,
//  and the tuple_iterator is the only structure that can store this state
//  information
// *************************************************************************
Cor_StorageRelation *Cor_do_query(Cor_QFModuleData *md,
    Cor_StorageRelation *magic,
    int old_magic_relation, Cor_TupleIterator *iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_Table *table_copy = NULL;
    Cor_ModuleInfo *temp_module_info;

#    ifdef DO_PROFILE
        md->module_info->Cor_GlobalProfileInfo.clear();
#    endif

    // If staggered evaluation is not being performed, or this is the
    // first time do_query is being called on this module for this query
    // initialize the table of relations
    if (!iterator || !iterator->state) {
        // Make a copy of the table.  This is to handle recursive calls
        // to do_query.  Each call should see an unmodified Table.
        table_copy = initialize_do_query(md, magic, old_magic_relation);
        if (!table_copy)
            return(NULL);
    } else {
        table_copy = iterator->state->table;
    }

    Cor_T_Stack->push(table_copy);

    // THIS SHOULD BE PART OF PUSH !!    PRAVEEN
    temp_module_info = Cor_CurrModuleInfo;
    Cor_CurrModuleInfo = table_copy->moduledata->module_info;

#    ifdef DEBUG
        if (Cor_exEnv->dbg_modules) {
            *(Cor_exEnv->trace_strm) << "Entering module\n\n";
        }
#    endif

    Cor_StorageRelation *answer_rel = (Cor_StorageRelation *)(table_copy->
        rels[Cor_get_offset(table_copy, md->export_pred)].relPtr);

    // For the case lazy_eval-, staggered evaluation is not executed.
    // Since md->Solve() is called with NULL as the iterator argument,
    // the module is completely evaluated, and the iterator->state
    // field remains NULL.
    if (!iterator || !(md->module_info->LazyEval)) {

        md->Solve(table_copy, NULL);
        Cor_cleanup_do_query(md, magic, table_copy);

          // free the top of the stack
        Cor_Table *tos = Cor_T_Stack->top();
        Cor_T_Stack->pop();
        delete tos;


        // THIS SHOULD BE PART OF PUSH !!    PRAVEEN
        Cor_CurrModuleInfo = temp_module_info;

        return answer_rel;
    } else {
        // if this is the first time the module is being called with the
        // iterator, the iterator->state field will be created.
        // otherwise, the existing iterator->state field will be used.
        // when there are no further answers, the state field is deleted and
        // set to NULL by SccInfo::Solve(), and this signifies that the
        // Table and temporary relations can be cleaned up.
        // this is the first invocation of the module using this iterator
        if (!(iterator->state)) {
            iterator->state = new Cor_ModuleExecInfo(md, table_copy, magic,
                Cor_do_query);
        }

        // this sets the rm_Arr field
        md->Solve(table_copy, iterator);

        // if no more answers
        if (!(iterator->state->rm_Arr)) {

            Cor_cleanup_do_query(iterator->state->md, iterator->state->magic,
                iterator->state->table);
              // NOTE that the state->scan_mark field is redundant !!
            delete iterator->state;
            iterator->state = NULL;

              // free the top of the stack
            Cor_Table *tos = Cor_T_Stack->top();
            Cor_T_Stack->pop();
            delete tos;

             // THIS SHOULD BE PART OF PUSH !!    PRAVEEN
            Cor_CurrModuleInfo = temp_module_info;

            return answer_rel;
        }

        // even for lazy evaluation, pop the table stack, or else,
        // the table stack abstraction is lost !
          // This table will be reused (don't delete it)
        Cor_T_Stack->pop();

          // THIS SHOULD BE PART OF PUSH !!    PRAVEEN
        Cor_CurrModuleInfo = temp_module_info;

        return answer_rel;
    }
}

// *************************************************************************
// Fill the Table with the appropriate relation pointers
// *************************************************************************
static void fill_table(Cor_Table *F)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for (int i = 0; i < F->numRels; i++) {
        F->rels[i].AssignRel(Cor_Find_External_Relation(F->rels[i].relName,
            F->rels[i].relArity));
        F->rels[i].relPtr->local_options = Cor_exEnv->GlobalRelationOptions;
    }
}

// *************************************************************************
// Executes a single rule from the parserStruct. This function is invoked
// from the routines in ParseActions.C to execute imperative rules.
// *************************************************************************
int Cor_execute_single_rule(Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!Cor_parserStruct->snrule) {
        Cor_parserStruct->snrule = new Cor_SemiNaive();

        // get the info out of the parserStruct, and create the RuleInfo
        // structure, relation Table and seminaive rule structure
        Cor_compile_single_rule(Cor_parserStruct, Cor_parserStruct->table,
            Cor_parserStruct->rInfo_arr, *(Cor_parserStruct->snrule));
     }

      // fill the table with the appropriate relation pointers
    fill_table(Cor_parserStruct->table);

    if (coralConfig->server_mode) {    // automatic suspension

      Cor_suspension->set = Cor_convert_ans_2_set(Cor_curUser->num_answers);
    }

      // evaluate the rule
    //cout << "In " << __FUNCTION__ << " about to call evaluate.\n";
    int new_tuples =
      Cor_parserStruct->snrule->evaluate(Cor_parserStruct->table,NULL,NULL,1);

      // Finally, its safe to delete magic_rel allocated in DerivedRelation.C
      // LEAK - This misses some for Lazy Evaluation
    if ( Cor_parserStruct->table->moduledata
                && Cor_parserStruct->table->moduledata->magic_rel ) {
        delete Cor_parserStruct->table->moduledata->magic_rel;
        Cor_parserStruct->table->moduledata->magic_rel = NULL;
    }

    return(new_tuples);
}

// *************************************************************************
