
//**************************************************************************
//    ConnectArray.C
//
//    The ConnectArray Class - Used to hold an array of connections. This
//    class is the top-level item in the implemenatation of the Coral
//    client/server stuff. A ConnectArray holds an array of Connections,
//    which each hold an array of Users.
//     
//**************************************************************************

#include "ConnectArray.h"
#include "Globals.h"

//**************************************************************************
// ConnectArray Constructor - Just clears everything.
//**************************************************************************
//Cor_ConnectArray::Cor_ConnectArray()
//{
//}

//**************************************************************************
// ConnectArray Destructor - Closes the server socket (if any), then closes
//    each connection in the array.
//**************************************************************************
Cor_ConnectArray::~Cor_ConnectArray()
{
}

//**************************************************************************
// ListWho() - 
//    This function lists all users for the connection specified, 
//    (printing the output to the file pointer given), if the connection
//    specified is NULL (the default value), lists all users on all
//    connections.
//
//    Return: The number of users on this (or all) connections.
//
//**************************************************************************
int Cor_ConnectArray::ListWho(CoralStream *strm, Cor_Connection *conn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int cnt = 0;

    if (conn){
        cnt += conn->ListWho(strm);
    } else {
        for (int i=0; i < MAX_CONNECTIONS; i++) {
            if (connArray[i])
                cnt += connArray[i]->ListWho(strm);
        }
    }

    return(cnt);
}

//**************************************************************************
int Cor_ConnectArray::LocalConnection()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

      // Must not have any other connections yet.
    if (num_conn == 0) {

        cur_conn = 0;

          // Make a local connection
        connArray[0] = new Cor_Connection(NULL);
        connArray[0]->conn_id = 0;

        num_conn = 1;

          // Start using the connection, user 'nobody'
        connArray[0]->useConnection();
        connArray[0]->useUser(NOBODY);
    }

    return(0);
}

//**************************************************************************
void Cor_ConnectArray::DeleteLocalConnection()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

      // Delete the local connection
    delete connArray[0];
    connArray[0] = (Cor_Connection *) NULL;

    num_conn--;

    Cor_exEnv = Cor_deflt_exEnv;
    Cor_parseStack = NULL;
    Cor_parseEnv = NULL;
}

//**************************************************************************
