// *************************************************************************
//    CRdbRelation.C
// *************************************************************************

#include "CRdbManager.h"

#ifdef DO_RDB

#include "CRdbRelation.h"
#include "Interface.h"
#include "Database.h"
#include "Globals.h"
#include "CoralError.h"
#include "SymTable.h"
#include "LogStream.h"
#include "BoundArgList.h"
#include "Log.h"
#include "String.h"
#include "Debug.h"


// *************************************************************************
// make this more global...
ostream& operator <<(ostream& os, const RdbItemList& items)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char sep = '(';
    for ( int i = 0 ; i < items.Size() ; i++ ) {
        os << sep << items[i];
        sep = ',';
    }
    return os << ')';
}

// *************************************************************************
// this should really be the constructor, & it could delete the
// previous version
int CreateCRdbRelation(Cor_CArg& newName, RdbJoin& join)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (Cor_CurDB->RelationTable->lookup(newName)) // name exists!
      return COR_BAD_ARGLIST;

    Cor_CRdbRelation *rel = new Cor_CRdbRelation(newName,join);

      // Insert the new relation in the relation table.
    Cor_CurDB->RelationTable->insert(newName, (Pointer) rel);
    return(0);
}

// *************************************************************************

Cor_CRdbRelation::Cor_CRdbRelation(const Cor_CArg& newName,
        const RdbJoin& newJoin)
        : Cor_Relation(), joinInfo(newJoin)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // register with the manager
    // pCRdbManager->StoreCRdbRelation(this);
    ASSERT( joinInfo.RelCount() > 0 );
    name = newName;
    r_kind = (RelationKind) COR_R_RDBREL;
    queryVersions = new RdbQueryVersions(joinInfo);
    _arity = queryVersions->Arity();
    rdbInsert = NULL;
    deleteVersions = NULL;
}

// *************************************************************************

Cor_CRdbRelation::~Cor_CRdbRelation()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if ( deleteVersions ) {
        delete deleteVersions;
        deleteVersions = NULL;
    }

    if ( rdbInsert ) {
        delete rdbInsert;
        rdbInsert = NULL;
    }

    if ( queryVersions ) {
        delete queryVersions;
        queryVersions = NULL;
    }

    // remove from manager
    //pCRdbManager->RemoveCRdbRelation(this);
}

// *************************************************************************

RdbCursor *Cor_CRdbRelation::Init(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    RdbCursor *csr = NULL;
    if ((iterator.ipos == Scanning_Init)) {
        csr = Bind( iterator );
        iterator.stack_mark.get_mark();
        iterator.ipos = Scanning_Known;
        iterator.scanDesc = csr;
    } else {
        iterator.stack_mark.pop_to();
        csr = (RdbCursor*)iterator.scanDesc;
    }
    return csr;
}

// *************************************************************************

Cor_Tuple *Cor_CRdbRelation::Done(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if ( iterator.scanDesc ) {               
        delete (RdbCursor*)iterator.scanDesc;
        iterator.scanDesc = NULL;
    }
    iterator.ipos = Scanning_Done;
    iterator.set_no_match();
    return NULL;
}

// *************************************************************************

Cor_Tuple *Cor_CRdbRelation::get_next_tuple(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (Cor_exEnv->C_interrupt_raised)
      return Done(iterator);

    RdbCursor *csr = Init(iterator);
    if ( !csr )
      return Done(iterator);

    RdbItemList result;

    // keep getting tuples from the rdbms until:
    //   a. there are no more (eof)
    //   b. there is an error
    //   c. the tuple unifies with the iterator - should this be done??
    while( !csr->GetNext(result) ) {

        LOG( "Rec returned: " << result << endl );

        // convert the RdbItemList to an ArgList
        Cor_ArgList *arglist = AsArgList(result);

        // if the tuple unifies, return it
        Cor_Tuple *tuple = new Cor_Tuple(arglist);
        if ( Cor_unify_literal(iterator,tuple,Cor_exEnv->dbg_get_next) )
          return tuple;

        // tuple did not unify, clean up and get another one.
        delete tuple;
        delete arglist;    
    }
    // eof or error
    return Done(iterator);
}

// *************************************************************************

RdbCursor *Cor_CRdbRelation::Bind(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    RdbItemList bindings;
    BitVector adorn;
    if ( AsRdbItemList(iterator.arg_list, iterator.bindenv, bindings, adorn) )
      return NULL;
    RdbQuery& query = queryVersions->Lookup(adorn);
    return new RdbCursor(query, bindings);
}

// *************************************************************************

// returns 0 on success, 1 on error
int Cor_CRdbRelation::AsRdbItemList(Cor_ArgList& argList,
                Cor_BindEnv *pBindEnv,
                RdbItemList& result,
                BitVector& adorn) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int arity = argList.arity();
    adorn.setlen(arity);    // adornment of bindings
    for ( int i = 0 ; i < arity ; i++ ) {
        Cor_Term term(argList[i], pBindEnv);
        Cor_deep_dereference_term(term);
        switch( term.expr.Kind() ) {
          case SYMBOL_ARG:
          case TINY_SYM_ARG:
            {
            adorn.set(i);
            char *str = term.expr.symbol();
            result.Append( RdbItem(str) );
            break;
            }
          case INT_ARG:
            adorn.set(i);
            result.Append(RdbItem(long(term.expr.NumIval())));
            break;
          case DOUBLE_ARG:
            adorn.set(i);
            result.Append(RdbItem(term.expr.NumDval()));
            break;
          case VAR_ARG:
            // do nothing
            break;
          default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "CRdbRelation::AsRdbItemList");
            return 1;
        }
    }
    return 0;
}

// *************************************************************************

Cor_ArgList *Cor_CRdbRelation::AsArgList(RdbItemList& result)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ArgList *arglist = new Cor_ArgList(result.Size());
    for ( int i = 0 ; i < result.Size() ; i++ ) {
        switch( result[i].Type() ) {
          case RdbItem::RDB_STRING:
            {
            String str = String(result[i]); 
            str.RTrim();
            Cor_CArg *carg = make_arg((const char*) str);
            (*arglist)[i] = *carg;
            carg->destroy();
            break;
            }
          case RdbItem::RDB_INT:
            (*arglist)[i] = Cor_CArg(int(RdbInt(result[i])));
            break;
          case RdbItem::RDB_REAL:
            (*arglist)[i] = Cor_CArg(double(RdbReal(result[i])));
            break;
          case RdbItem::RDB_NULL:
            {
            char *na = "(null)";
            // char *s = new char[ strlen(na) +1 ];
            // strcpy(s,na);
            Cor_CArg *carg = make_arg(na);
            (*arglist)[i] = *carg;
            carg->destroy();
            break;
            }
          default:
            {
            ASSERT(0);
            }
        }
    }
    return arglist;
}

// *************************************************************************

void Cor_CRdbRelation::release( Cor_TupleIterator& iterator )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Done(iterator);
}

// *************************************************************************

int Cor_CRdbRelation::insert_tuple(Cor_Tuple *newTuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_BindEnv *env = NULL;
    return insert_new(newTuple->args(), env);
}

// *************************************************************************

// check subsumption, then inserts into backend dbms
int Cor_CRdbRelation::insert_new(Cor_ArgList& argList, Cor_BindEnv *pBindEnv, 
                 Cor_BindEnv* , Cor_Tuple* , int )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (Cor_exEnv->C_interrupt_raised) return 0;  

    TRACE_INSERTION("[Starting to insert " << name.symbol()
            << '(' << corBoundArgList(argList,pBindEnv) << ')' );

    if ( joinInfo.RelCount() != 1 ) {
      ERROR_MSG(" - attempt to insert tuple into a join of rdb relations\n");
      TRACE_INSERTION("insertion failed]\n");
      return 0;
    }

    if ( !rdbInsert ) {        // allocate if first try to insert
        rdbInsert = joinInfo.Rel(0).NewInsert();
    }

    if ( argList.arity() != arity() ) {
        ERROR_MSG("attempt to insert tuple of arity " << argList.arity()
              << "\n\t\t- in relation " << name.symbol()
              << " of arity " <<  arity() << endl);
        TRACE_INSERTION("insertion failed]\n");
        return 0;
    }

    RdbItemList tuple;
    BitVector adorn(arity());

    if ( AsRdbItemList(argList, pBindEnv, tuple, adorn) ) {
        // error reported in AsRdbItemList
        TRACE_INSERTION("insertion failed]\n");
        return 0;
    }

    if ( adorn.count() != arity() ) { // all of the args must be bound
        ERROR_MSG("attempt to insert a non-ground tuple\n");
        TRACE_INSERTION("insertion failed]\n");
        return 0;
    }
    
    // subsumption (duplicate) checks
    if (check_subsum==1) {
        RdbQuery& qry = queryVersions->Lookup(adorn); // fully bound
        RdbCursor csr(qry, tuple);
        RdbItemList result;
        if ( !csr.GetNext(result) ) { // tuple found, subsume
            // this is not an error
            TRACE_INSERTION("-- subsumed]\n");
            return 0;
        }
    }

    int rc = rdbInsert->Insert(tuple);
    if ( rc ) {
        ERROR_MSG("received rdb error " << rc << endl);
        TRACE_INSERTION("insertion failed]\n");
        return 0;
    }

    TRACE_INSERTION("-- done]\n");
    return 1;            // successfully added a new row
}

// *************************************************************************

// delete all tuples that match 'tuple'
int Cor_CRdbRelation::tuple_delete(Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (Cor_exEnv->C_interrupt_raised) return 0;  
    
    if ( joinInfo.RelCount() != 1 ) {
        ERROR_MSG("attempt to delete tuple into a join of rdb relations\n");
        return 0;
    }

    if ( !deleteVersions ) {    // allocate if this is the first time
        deleteVersions = new RdbDeleteVersions(joinInfo.Rel(0));
    }

    RdbItemList rdbTuple;
    BitVector adorn(tuple->arity());
    if ( AsRdbItemList(tuple->args(), NULL, rdbTuple, adorn) ) {
        // error already reported
        return 0;
    }

    RdbDelete& del = deleteVersions->Lookup(adorn);
    RdbError::ErrorCode rc = del.Delete(rdbTuple);
    if ( rc == RdbError::SQL_EOF ) {
        // no matching rows found!
        return 0;
    }

    if ( rc ) {
        ERROR_MSG("received rdb error " << rc << endl);
        return 0;
    }

    return 1;            // successfully added a new row
}

// *************************************************************************

void Cor_CRdbRelation::print_name( ostream *out )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    String sys = queryVersions->SystemName();
    String db = queryVersions->DatabaseName();
    String schema = queryVersions->Schema();

    *out << name.symbol() << "/" << arity() << "\t: "
         << (const char*)sys << "/"
         << (const char*)db << ":"
         << (const char*)schema << endl;
}

// *************************************************************************

// takes a funcarg like: coralname(rel1(X,Y),rel2(Y,Z)...)
int CreateCRdbJoinRelation(Cor_CArg *pfaNewRel, Cor_BindEnv *pBindEnv)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int relCount = pfaNewRel->arity();
    if (relCount < 1)        // must select at least one rel
      return COR_BAD_ARGLIST;

    RdbRelList relList;
    RdbJoinCondList joinList;
    Cor_CRdbVarPtrList varList;

    // foreach relation to be joined
    for (int rel = 0; rel < relCount; rel++ ) {
        Cor_Term tRel((pfaNewRel->args())[rel], pBindEnv);
        Cor_deep_dereference_term(tRel);
        if (! tRel.expr.is_functor() )
          return COR_BAD_ARGLIST;
        
        Cor_CArg *pfaRel = &(tRel.expr);
        int arity = pfaRel->args().arity();
        if (arity < 1)        // must select at least one attr.
          return COR_BAD_ARGLIST;
    
        Cor_Relation *pRel = find_relation(Cor_CArg(pfaRel->functor()),arity);
        if ( pRel == NULL )    // couldn't find relation
          return COR_BAD_ARGLIST;
        
        if ( pRel->r_kind != COR_R_RDBREL ) // not an RDB relation
          return COR_BAD_ARGLIST;
    
        // make ptr to crdb subtype
        Cor_CRdbRelation *pCRdbRel = (Cor_CRdbRelation*)pRel; 
    
        // need to check that all are from the same system...
        // or should Rdb__ be responsible, probably.
    
        // should generalize this to work for joins too
        relList.Append(&pCRdbRel->RdbRel());
    
        // foreach attribute of this relation
        for ( int attr = 0 ; attr < arity ; attr++ ) {
            Cor_Term tAttr((pfaRel->args())[attr], pBindEnv);
            Cor_deep_dereference_term(tAttr);
            if ( tAttr.expr.is_var() ) {
                  // why are don't cares getting in here??
                Cor_CArg varName = Cor_CArg(tAttr.expr.symbol());
                if (! varName.equals("_")) { // skip don't cares
                    varList.Append(new Cor_CRdbVar(rel,attr,varName));
                }
            }
        }
    }
    CreateCRdbJoinList(joinList, varList);

    return CreateCRdbRelation(Cor_CArg(pfaNewRel->functor()),
            RdbJoin(relList,joinList));
}

// *************************************************************************

void CreateCRdbJoinList(RdbJoinCondList& joinList, Cor_CRdbVarPtrList& varList)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    varList.SortByName();
    int count = varList.Size();
    for ( int i = 1 ; i < count ; i++ ) {
        int j = i -1;
        while( j >= 0 && varList[i]->name == varList[j]->name ) {
            joinList.Append( RdbJoinCondition(varList[i]->rel,
                        varList[i]->col, varList[j]->rel, varList[j]->col));
            j--;
        }
    }
}

// *************************************************************************

// implements insertion sort to put the list in order by name
void Cor_CRdbVarPtrList::SortByName()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CRdbVar *t;
    int count = Size();
    for ( int i = 1 ; i < count ; i++ ) {
      t = (*this)[i];
      int j = i;
      while((j > 0)
          && ( strcmp((*this)[j-1]->name.symbol(), t->name.symbol()) >0) ) {
        Set(j,(*this)[j-1]); // want just a ptr swap, not struct
        j--;
      }
      Set(j,t);
    }
}

// *************************************************************************

#endif // DO_RDB
