// *************************************************************************
//    CoralError.C
// *************************************************************************

#include "CoralError.h"
#include "Globals.h"
#include "Debug.h"

// *************************************************************************

static char *Error_Table[ERROR_TABLE_SIZE] = {
    "",                    // 0
    "syntax error",                // COR_BAD_SYNTAX       -1
    "unknown annotation",                // COR_BAD_ANNO        -2
    "quit inside module",            // COR_QUIT_IN_MODULE   -3
    "not implemented" ,            // COR_NOT_IMPLEMENTED  -4
    "illegal insert of tuple into relation",// COR_BAD_TUPLE_INSERT -5
    "illegal call of empty_relation()",     // COR_BAD_EMPTY_REL    -6
    "bad argument list",            // COR_BAD_ARGLIST    -7
    "bad argument type",            // COR_BAD_ARG_TYPE    -8
    "non-numeric arguments passed to builtin relation", 
                                // COR_NON_NUMERIC_ARGS -9
    "comparison of two non-numbers" ,    // COR_NON_NUMERIC_COMP -10
    "wrong arity",                // COR_BAD_ARITY        -11
    "not a valid relation name",        // COR_BAD_REL_NAME     -12
    "unknown global variable",        // COR_BAD_GLOBAL_VAR    -13
    "empty argument list",            // COR_EMPTY_ARGLIST    -14
    "argument should be a string",        // COR_NON_STRING_ARG   -15
    "cannot find file",            // COR_FILE_NOT_FOUND   -16
    "failed to execute command",        // COR_FAILED_COMMAND   -17
    "nested set is not ground",        // COR_NESTED_SET_NOT_GROUND-18
    "bad schema definition",        // COR_BAD_SCHEMA    -19
    "too many arguments",            // COR_TOO_MANY_ARGS    -20
    "error in index",            // COR_INDEX_ERR    -21
    "bad database tuple",            // COR_BAD_TUPLE    -22
    "Exodus error",                // COR_ESM_ERR        -23
    "multiple ws error",            // COR_MULTIPLE_DB_ERR    -24
    "grouping operator in rule body",    // COR_GROUPING_ON_RHS    -25
    "module not found",            // COR_MODULE_NOT_FOUND -26
    "backend rdbms error",            // COR_RDB_ERR          -27
    "exhausted free store !!",        // COR_NO_MEMORY        -28
    "Fatal Coral Error",              // COR_FATAL_ERROR      -29
    "",                    //                      -30
    "",                    //                      -31
    "",                    //                      -32
    "",                    //                      -33
    "",                    //                      -34
    "",                    //                      -35
    "",                    //                      -36
    "",                    //                      -37
    "",                    //                      -38
    "",                    //                      -39
    "",                    //                      -40
    "",                    //                      -41
    "",                    //                      -42
    "",                    //                      -43
    "",                    //                      -44
    "",                    //                      -45
    "",                    //                      -46
    "",                    //                      -47
    "",                    //                      -48

    "internal system error --- please notify software supplier",
                                // COR_INTERNAL_ERROR    WMI-1

                        // WARNING_MESSAGE_INDEX = 50

    "unable to find relation",        // COR_REL_NOT_FOUND    WMI
    "cannot open file",            // COR_CANNOT_OPEN_FILE  +1
    "received signal signifying exception",    // COR_SIGNAL_RECEIVED   +2
    "duplicate schema defintion",        // COR_DUP_SCHEMA        +3 
    "" ,
    "" ,
    "" ,
    "" ,
    "" ,
    "" ,
    "" ,
    "" ,
    "" ,
    "" ,
    "" ,
    "" ,
    "" ,
    "" ,
    "" ,
    "" ,
    "" ,
    "" ,
    "" ,
    "" ,
    ""  
    // ERROR_TABLE_SIZE 
};

// *************************************************************************
//    Print error message, followed by optional message and optional function
//    name.  Error codes are defined in cor_error.h.  The constant
//    WARNING_MESSAGE_INDEX must be properly defined.

void CORAL_error(int err_code, char *msg, char *fn_name,
     char *file, int line_no)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    DEBUG
    if (file)
        *(Cor_exEnv->err_strm) << 
            "In file '" <<  file << "' at line " <<  line_no << ":\n";
#endif
    err_code = -err_code;
    if (err_code >= WARNING_MESSAGE_INDEX)
        *(Cor_exEnv->err_strm) << "\nCORAL :: Warning --- ";
    else
        *(Cor_exEnv->err_strm) << "\nCORAL :: Error --- ";
 
    if ((err_code < ERROR_TABLE_SIZE)&&(err_code > -1))
        *(Cor_exEnv->err_strm) << " " <<  Error_Table[err_code];
    if (msg != NULL)
        *(Cor_exEnv->err_strm) << ": " << msg;
    *(Cor_exEnv->err_strm) << endl;
    if (fn_name && !Cor_exEnv->C_quiet_mode_default)
        *(Cor_exEnv->err_strm) << 
            "Warning/error flagged in function " << fn_name << "()\n";
}

// *************************************************************************

