// *************************************************************************
//    StorageManager.h
// *************************************************************************

#ifndef _StorageManager_H_
#define _StorageManager_H_

// *************************************************************************

class Cor_StorageManager {

  private:
    Cor_Catalog catalog;

    // Maybe this should have two buffer groups -- an MRU and an LRU
    Cor_BufferGroup buf;
    Cor_Transaction trans;
    VOLID _volid;
    VOLID _temp_volid;

  public:
    Cor_StorageManager()    { INVALIDATE_VOLID(_volid); }
   ~Cor_StorageManager()    { shutdown(); }
    void init();

    void shutdown();
    int bufGroup()        { return buf.bufGroup; }
    VOLID volid()        { return _volid; }
    VOLID tempvolid()    { return _temp_volid; }

    // Transaction handling functions
    void startTrans()    { trans.start() ; }
    int commitTrans()    { return trans.commit() ;}
    void abortTrans()    { trans.abort() ;}

    // Catalog handling functions

    // Add a file to the catalog
    int addFile(Cor_CArg& name, int arity, Cor_FileDescStruct *obj);

    // Check if a file is in the catalog
    int lookupFile(Cor_CArg& name, int arity, OID *oid);

    // Delete a file from the catalog
    int deleteFile(Cor_CArg& name, int arity);

    // List all files/schemas
    void listFiles(ostream *outf);

    // Object handling functions

    // Retrieve an object specified by oid. Note that the object is PINNED
    // in the buffer pool !
    int retrieveObject(OID *oid, USERDESC ** u_desc);

    // Release (unpin) an object from the buffer pool.
    int releaseObject(USERDESC *u_desc);

    // Destroy an object. If bufGroup is 0, the object is in the catalog file.
    int destroyObject(OID *oid, int bufGroup);
};

// *************************************************************************

extern Cor_StorageManager SM;

// *************************************************************************

#endif _StorageManager_H_
