
//**************************************************************************
//    WodiStream.h
//
//    The Cor_wodibuf Class - A buffer class for providing stream access
//    to Wodi communication. 
//
//**************************************************************************

#ifdef DO_WODI

#ifndef _WODISTREAM_H_
#define _WODISTREAM_H_


#include <string.h>
#include <errno.h>
#include <fstream.h>
#include "CoralStream.h"
#include "pvm3.h"
#include "wodi.h"


// msg tags used by wodi (magic integers)
#define CORAL_RESP_TAG 23458
#define CORAL_WORK_TAG 23459

enum BUF_TYPE {WODI_MASTER, WODI_SLAVE};


class Cor_wodibuf : public streambuf {

public:
    int wodi_id;

private:

    BUF_TYPE type;              
    
    int get_buffer_size;        // the buffer size of the get area
    int put_buffer_size;        // the buffer size of the put area

    Cor_wodibuf(const Cor_wodibuf& other);
    Cor_wodibuf& operator=(const Cor_wodibuf& other);

    char * backup;              // the backup of the query 

    int total_len;
    
    ofstream *w_out, *w_err;    // for redirecting output 
    
protected:

    // consume characters in the put area
    int overflow(int c = EOF);

    // supply characters for fetching. put them into the get area 
    int underflow();          

    // allocate the put and get areas
    int doallocate();

    // initialize the buffer
    void  init();

public:

    // the constructor 
    Cor_wodibuf(BUF_TYPE _type);

    // the destructor
    ~Cor_wodibuf();

    // check if the Cor_wodibuf is valid
    int Valid();
    
    // synchronize the put area. 
    int sync();	

    int wodi_flush();

};

//**************************************************************************


#endif   // _WODISTREAM_H_

#endif // DO_WODI
