// *************************************************************************
//    SipOrder.C
// *************************************************************************

#include "SipOrder.h"

#include "Config.h"
#include "Rule.h"
#include "Assert.h"
#include "Sips.h"
#include "Unparse.h"
#include "Globals.h"

// *************************************************************************

static int torder[MAX_RHSPREDS];

static void toporder(Cor_Clause *rule, int torder[]);

// *************************************************************************
/*
    graph(i,j) = 1 if there is a sip arc from pred[i] to pred[j] 
        (i = 0 implies the head of the rule )
*/

static void toporder(Cor_Clause *rule, int torder[])
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    static char    graph[MAX_RHSPREDS+1][MAX_RHSPREDS+1];
    static int countarr[MAX_RHSPREDS+1];
    static int noarcs[MAX_RHSPREDS+1];
    int    i, j, k, count;
    Cor_Sip *s;
    int    b;

    ASSERT(rule->num_preds <= MAX_RHSPREDS);
    for(i=0; i <= rule->num_preds; i++)
        countarr[i] = 0;

    // Initialize graph.
    for (i=0; i<=rule->num_preds; ++i) {
        for (j=0; j<=rule->num_preds; ++j)
            graph[i][j] = 0;
        noarcs[i] = 1;
    }

    //    Construct graph from sips ans set up count array.
    for (s = rule->sips->sip_list; s; s = s->next) {
        for (b=firstpredinsip(&s->tail); b>=0; b=nextpredinsip(&s->tail,b)) {
            graph[b][s->head] = 1;
            noarcs[b] = noarcs[s->head] = 0;
        }
    }

       for (i=0; i<=rule->num_preds; i++) {
           for (j=0; j<=rule->num_preds; j++) {
            if ( graph[i][j] ) countarr[j]++;
        }
    }

    // Force all preds with no incoming and no outgoing sip arcs to the 
    // end of the rule by giving them a count of one.
    for (i=1; i <= rule->num_preds; i++)
        if (noarcs[i]) countarr[i] = 1;

    count = 0;
    for (k=0; k<=rule->num_preds; ++k) {
        for (i=0; i<=rule->num_preds; ++i) {
            if (countarr[i] == 0) {
                // Mark as done.
                countarr[i] = -1;
                if (i > 0) torder[i-1] = count++;
                // Bump down the pred numbering by 1 for torder to start
                // body pred numbering at 0.
                for (j=1; j <= rule->num_preds; j++) {
                    if (graph[i][j]) {
                        countarr[j]--;
                        ASSERT( countarr[j] >= 0 );
                    }
                }
            }
        }
    }

    // Set up rule->last_part properly, and make sure the order of
    // predicates in the rule body is such that anything beyond last_part
    // does not participate in any sip.

    rule->last_part = count -1; 
    // Now tack on preds that had no sip arcs entering or leaving.
    for (i=1; i <= rule->num_preds; i++) {
        if (noarcs[i]) {
            ASSERT(countarr[i] == 1);
            torder[i-1] = count++;
        }
    }


#    ifdef DEBUG
        if (0) {
            *cout_strm << *rule << "torder array:";
            for (i = 0; i < rule->num_preds; i++)
                *cout_strm << i << " ";
            *cout_strm << endl;
        }
#    endif
}

/*****************************************************************************/
//
//  Reorder the right-hand sides of the adorned system so the predicate
//  occurrences follow the order induced by the sips.
//

void siporder(Cor_Collection *ad_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i, b;
    Cor_Sip *s;
    Cor_Sips *sp;
    Cor_Literal *newpreds[MAX_RHSPREDS];
    Cor_Clause *adrule;
    int j, nrules;

    nrules = ad_rules->count();

    for (j=0; j < nrules; j++) {
        adrule = ad_rules->get_ith(j);

        toporder(adrule, torder);

        // Rearrange right-hand side.
        for (i=0; i<adrule->num_preds; ++i) {
            newpreds[torder[i]] = adrule->preds[i];
        }
        for (i=0; i<adrule->num_preds; ++i) {
            adrule->preds[i] = newpreds[i];
        }

        //    Fix sips to reflect rearrangement.
        for (sp = adrule->sips; sp; sp = sp->next) {
            for (s = sp->sip_list; s; s = s->next) {
                Cor_BitVector bitvec(s->tail.len());
                for (b = firstpredinsip(&s->tail); b>=0;
                    b = nextpredinsip(&s->tail, b))
                {
                    if (b == 0) {
                        //    Head predicate
                        bitvec.set(0);
                    }
                    else {
                        // Add 1 since 0 is head
                        bitvec.set(torder[b-1]+1); 
                    }
                }
                s->tail = bitvec;
                s->head = torder[s->head-1]+1;
                
            }
        }

#    ifdef DEBUG
        if (0) {
            *cout_strm << "Modified original rule:\n" << *adrule << endl;
        }
#    endif
    }
}

// *************************************************************************
//
//  Find a sip for a predicate in an adorned rule.  Non-matching sips have
//  already been discarded.
//
Cor_Sip *findadsip(Cor_Clause *adrule, int p)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Sip *s;

    for (s=adrule->sips->sip_list; s; s=s->next) {
        // In sip, body pred numbering starts from 1.
        if (p == s->head-1)
            break; 
    }
    return(s);
}

// *************************************************************************
